@@ -10,9 +10,10 @@
 #include "fail.hpp"  // for fail
 #include "util.hpp"  // for starts_with, to_lower
 #include "tostr.hpp"  // for tostr
+#include <cassert>
+#include <cstring>   // for memchr
 #include <algorithm> // for move, find_if, all_of, min, rotate
 #include <array>
-#include <cstring>   // for memchr
 #include <initializer_list>
 #include <iosfwd>    // for size_t, ptrdiff_t
 #include <new>
@@ -297,6 +298,7 @@ struct Table {
   bool ok() const { return !positions.empty(); }
   size_t width() const { return positions.size(); }
   size_t length() const;
+  size_t size() const { return length(); }
   bool has_column(int n) const { return ok() && positions.at(n) >= 0; }
   int first_of(int n1, int n2) const { return positions.at(n1) >= 0 ? n1 : n2; }
   Row tags() { return Row{*this, -1}; }
@@ -330,6 +332,8 @@ struct Table {
   void append_row(std::initializer_list<std::string> new_values) {
     append_row<std::initializer_list<std::string>>(new_values);
   }
+  void remove_row(int row_index) { remove_rows(row_index, row_index+1); }
+  void remove_rows(int start, int end);
   Column column_at_pos(int pos);
   Column column(int n) {
     int pos = positions.at(n);
@@ -355,6 +359,8 @@ struct Table {
 
   void erase();
 
+  void convert_pair_to_loop();
+
   // It is not a proper input iterator, but just enough for using range-for.
   struct iterator {
     Table& parent;
@@ -617,7 +623,7 @@ inline Table::Row Table::find_row(const std::string& s) {
   } else if (as_string(bloc.items[pos].pair[1]) == s) {
     return Row{*this, 0};
   }
-  fail("Not found in the first column: " + s);
+  fail("Not found in " + *column_at_pos(pos).get_tag() + ": " + s);
 }
 
 template <typename T> void Table::append_row(T new_values) {
@@ -626,8 +632,7 @@ template <typename T> void Table::append_row(T new_values) {
   if (new_values.size() != width())
     fail("append_row(): wrong row length");
   if (!loop_item)
-    // this limitation could be lifted if needed
-    fail("append_row(): existing data must be in loop_");
+    convert_pair_to_loop();
   Loop& loop = loop_item->loop;
   size_t cur_size = loop.values.size();
   loop.values.resize(cur_size + loop.width(), ".");
@@ -636,6 +641,20 @@ template <typename T> void Table::append_row(T new_values) {
     loop.values[cur_size + positions[n++]] = value;
 }
 
+inline void Table::remove_rows(int start, int end) {
+  if (!ok())
+    // this function is used mostly through remove_row()
+    fail("remove_row(): table not found");
+  if (!loop_item)
+    convert_pair_to_loop();
+  Loop& loop = loop_item->loop;
+  size_t start_pos = start * loop.width();
+  size_t end_pos = end * loop.width();
+  if (start_pos >= end_pos || end_pos > loop.values.size())
+    throw std::out_of_range("remove_row(): invalid index");
+  loop.values.erase(loop.values.begin() + start_pos,
+                    loop.values.begin() + end_pos);
+}
 
 inline Column Table::column_at_pos(int pos) {
   if (loop_item)
@@ -651,6 +670,21 @@ inline void Table::erase() {
       bloc.items[pos].erase();
 }
 
+inline void Table::convert_pair_to_loop() {
+  assert(loop_item == nullptr);
+  Item new_item(LoopArg{});
+  new_item.loop.tags.resize(positions.size());
+  new_item.loop.values.resize(positions.size());
+  for (size_t i = 0; i != positions.size(); ++i) {
+    Item& item = bloc.items[positions[i]];
+    new_item.loop.tags[i].swap(item.pair[0]);
+    new_item.loop.values[i].swap(item.pair[1]);
+    item.erase();
+  }
+  loop_item = &bloc.items.at(positions[0]);
+  loop_item->set_value(std::move(new_item));
+}
+
 inline const Item* Block::find_pair_item(const std::string& tag) const {
   for (const Item& i : items)
     if (i.type == ItemType::Pair && i.pair[0] == tag)
@@ -718,7 +752,7 @@ inline size_t Block::get_index(const std::string& tag) const {
         (item.type == ItemType::Loop && item.loop.find_tag(tag) != -1))
       return i;
   }
-  fail(tag + "is not in block");
+  fail(tag + " not found in block");
 }
 
 inline void Block::move_item(int old_pos, int new_pos) {
