@@ -6,16 +6,12 @@
 #include "gemmi/to_json.hpp"
 #include "gemmi/fstream.hpp"
 
-//#include <pybind11/pybind11.h>
+#include "common.h"
 #include <pybind11/stl.h>
 
 namespace py = pybind11;
 using namespace gemmi::cif;
 
-// defined in read.cpp
-void cif_parse_string(Document& doc, const std::string& data);
-void cif_parse_file(Document& doc, const std::string& filename);
-
 std::string pyobject_to_string(py::handle handle, bool raw) {
   PyObject* ptr = handle.ptr();
   if (ptr == Py_None) {
@@ -50,13 +46,21 @@ T& add_to_vector(std::vector<T>& vec, const T& new_item, int pos) {
   return vec[pos];
 }
 
+// for delitem_slice
+namespace gemmi { namespace cif {
+void delitem_at_index(Table& t, py::ssize_t idx) { t.remove_row((int)idx); }
+void delitem_range(Table& t, py::ssize_t start, py::ssize_t end) {
+  t.remove_rows((int)start, (int)end);
+}
+} } // namespace gemmi::cif
+
 void add_cif(py::module& cif) {
-  py::class_<Block> pyCifBlock(cif, "Block");
-  py::class_<Item> pyCifItem(cif, "Item");
-  py::class_<Loop> lp(cif, "Loop");
-  py::class_<Column> pyCifColumn(cif, "Column");
-  py::class_<Table> lt(cif, "Table");
-  py::class_<Table::Row> pyCifTableRow(lt, "Row");
+  py::class_<Block> cif_block(cif, "Block");
+  py::class_<Item> cif_item(cif, "Item");
+  py::class_<Loop> cif_loop(cif, "Loop");
+  py::class_<Column> cif_column(cif, "Column");
+  py::class_<Table> cif_table(cif, "Table");
+  py::class_<Table::Row> cif_table_row(cif_table, "Row");
 
   py::enum_<Style>(cif, "Style")
     .value("Simple", Style::Simple)
@@ -81,14 +85,8 @@ void add_cif(py::module& cif) {
         return d.blocks.at(index >= 0 ? index : index + d.blocks.size());
     }, py::arg("index"), py::return_value_policy::reference_internal)
     .def("__getitem__", [](Document &d, py::slice slice) -> py::list {
-        size_t start, stop, step, slicelength;
-        if (!slice.compute(d.blocks.size(), &start, &stop, &step, &slicelength))
-          throw py::error_already_set();
-        py::list l;
-        for (size_t i = 0; i < slicelength; ++i)
-          l.append(py::cast(d.blocks[start + i * step]));
-        return l;
-    })
+        return getitem_slice(d.blocks, slice);
+    }, py::return_value_policy::reference_internal)
     .def("__delitem__", [](Document &d, int index) {
         if (index < 0)
           index += (int) d.blocks.size();
@@ -147,7 +145,7 @@ void add_cif(py::module& cif) {
         return s;
     });
 
-  pyCifBlock
+  cif_block
     .def(py::init<const std::string &>())
     .def_readwrite("name", &Block::name)
     .def("__iter__", [](Block& self) { return py::make_iterator(self.items); },
@@ -241,12 +239,23 @@ void add_cif(py::module& cif) {
            }
            return data;
          }, py::arg("name"), py::arg("raw")=false)
+    .def("write_file",
+         [](const Block& self, const std::string& filename, Style s) {
+        gemmi::Ofstream os(filename);
+        write_cif_block_to_stream(os.ref(), self, s);
+    }, py::arg("filename"), py::arg("style")=Style::Simple,
+    "Write data to a CIF file.")
+    .def("as_string", [](const Block& self, Style style) {
+        std::ostringstream os;
+        write_cif_block_to_stream(os, self, style);
+        return os.str();
+    }, py::arg("style")=Style::Simple, "Returns a string in CIF format.")
     .def("__repr__", [](const Block &self) {
         return gemmi::tostr("<gemmi.cif.Block ", self.name, '>');
     });
 
 
-  pyCifItem
+  cif_item
     .def("erase", &Item::erase)
     .def_readonly("line_number", &Item::line_number)
     .def_property_readonly("pair", [](Item& self) {
@@ -260,7 +269,8 @@ void add_cif(py::module& cif) {
     }, py::return_value_policy::reference_internal)
     ;
 
-  lp.def(py::init<>())
+  cif_loop
+    .def(py::init<>())
     .def("width", &Loop::width, "Returns number of columns")
     .def("length", &Loop::length, "Returns number of rows")
     .def_readonly("tags", &Loop::tags)
@@ -275,7 +285,7 @@ void add_cif(py::module& cif) {
     });
 
 
-  pyCifColumn
+  cif_column
     .def(py::init<>())
     .def("get_loop", &Column::get_loop,
          py::return_value_policy::reference_internal)
@@ -301,7 +311,8 @@ void add_cif(py::module& cif) {
         return "<gemmi.cif.Column " + desc + ">";
     });
 
-  lt.def("width", &Table::width)
+  cif_table
+    .def("width", &Table::width)
     .def_readonly("prefix_length", &Table::prefix_length)
     .def_property_readonly("loop", &Table::get_loop,
                            py::return_value_policy::reference_internal)
@@ -314,12 +325,15 @@ void add_cif(py::module& cif) {
     .def("erase", &Table::erase)
     .def("append_row", &Table::append_row<std::vector<std::string>>,
          py::arg("new_values"))
+    .def("remove_row", &Table::remove_row, py::arg("row_index"))
     .def_property_readonly("tags",
             py::cpp_function(&Table::tags, py::keep_alive<0, 1>()))
     .def("__iter__", [](Table& self) {
         return py::make_iterator(self, py::keep_alive<0, 1>());
     }, py::keep_alive<0, 1>())
     .def("__getitem__", &Table::at, py::keep_alive<0, 1>())
+    .def("__delitem__", &Table::remove_row)
+    .def("__delitem__", &delitem_slice<Table>)
     .def("__bool__", &Table::ok)
     .def("__len__", &Table::length)
     .def("__repr__", [](const Table& self) {
@@ -329,7 +343,8 @@ void add_cif(py::module& cif) {
                ">";
     });
 
-  pyCifTableRow
+  cif_table_row
+    .def_readonly("row_index", &Table::Row::row_index)
     .def("str", &Table::Row::str)
     .def("__len__", &Table::Row::size)
     .def("__getitem__", (std::string& (Table::Row::*)(int)) &Table::Row::at)
