@@ -1,23 +1,15 @@
 // Copyright 2017 Global Phasing Ltd.
 
-#include "gemmi/align.hpp"     // for assign_label_seq_id
-#include "gemmi/elem.hpp"
-#include "gemmi/entstr.hpp"
-#include "gemmi/resinfo.hpp"
 #include "gemmi/model.hpp"
-#include "gemmi/calculate.hpp"
-#include "gemmi/polyheur.hpp"
-#include "gemmi/assembly.hpp"
-#include "gemmi/to_pdb.hpp"
-#include "gemmi/to_mmcif.hpp"
+#include "gemmi/calculate.hpp"  // for calculate_mass, count_atom_sites
+#include "gemmi/polyheur.hpp"   // for one_letter_code, trim_to_alanine
+#include "gemmi/assembly.hpp"   // for expand_ncs, HowToNameCopiedChain
 #include "gemmi/tostr.hpp"
-#include "gemmi/fstream.hpp"
 
-#include <fstream>
-#include <pybind11/pybind11.h>
+#include "common.h"
 #include <pybind11/stl.h>
 #include <pybind11/stl_bind.h>
-#include "common.h"  // for normalize_index
+#include "meta.h"
 
 namespace py = pybind11;
 using namespace gemmi;
@@ -25,25 +17,9 @@ using namespace gemmi;
 PYBIND11_MAKE_OPAQUE(std::vector<Connection>)
 PYBIND11_MAKE_OPAQUE(std::vector<NcsOp>)
 PYBIND11_MAKE_OPAQUE(std::vector<Entity>)
-PYBIND11_MAKE_OPAQUE(std::vector<Assembly>)
 using info_map_type = std::map<std::string, std::string>;
 PYBIND11_MAKE_OPAQUE(info_map_type)
 
-namespace gemmi {
-  inline std::ostream& operator<< (std::ostream& os, const Entity& ent) {
-    os << "<gemmi.Entity '" << ent.name << "' "
-       << entity_type_to_string(ent.entity_type);
-    if (ent.polymer_type != PolymerType::Unknown)
-      os << ' ' << polymer_type_to_qstring(ent.polymer_type);
-    os << " object at " << (void*)&ent << '>';
-    return os;
-  }
-}
-
-namespace pybind11 { namespace detail {
-  template<> struct type_caster<SeqId::OptionalNum>
-    : optional_caster<SeqId::OptionalNum> {};
-}} // namespace pybind11::detail
 
 template<typename T, typename C>
 C& add_item(T& container, C child, int pos) {
@@ -68,58 +44,32 @@ template<typename P> void remove_child(P& parent, int index) {
 }
 
 template<typename P> void remove_children(P& parent, py::slice slice) {
-  ssize_t start, stop, step, slength;
-  auto& children = parent.children();
-  if (!slice.compute(children.size(), &start, &stop, &step, &slength))
-    throw py::error_already_set();
-  for (int i = 0; i < slength; ++i) {
-    ssize_t idx = start + (step > 0 ? slength - 1 - i : i) * step;
-    children.erase(children.begin() + idx);
-  }
-}
-
-static std::vector<std::string>
-expand_protein_one_letter_string(const std::string& s) {
-  std::vector<std::string> r;
-  r.reserve(s.size());
-  for (char c : s)
-    r.push_back(expand_protein_one_letter(c));
-  return r;
+  delitem_slice(parent.children(), slice);
 }
 
 void add_mol(py::module& m) {
 
   // "Forward declaration" of python classes to avoid
   // C++ signatures in docstrings
-  py::class_<AtomAddress> pyAtomAddress(m, "AtomAddress");
-  py::class_<ResidueId> pyResidueId(m, "ResidueId");
   py::class_<Atom> pyAtom(m, "Atom");
   py::class_<Residue, ResidueId> pyResidue(m, "Residue");
   py::class_<Chain> pyChain(m, "Chain");
   py::class_<Model> pyModel(m, "Model");
-  py::class_<SeqId> pySeqId(m, "SeqId");
   py::class_<ResidueSpan> pyResidueSpan(m, "ResidueSpan");
   py::class_<ResidueGroup, ResidueSpan> pyResidueGroup(m, "ResidueGroup");
   py::class_<CraProxy> pyCraGenerator(m, "CraGenerator");
 
-  py::class_<ResidueInfo>(m, "ResidueInfo")
-    .def_readonly("one_letter_code", &ResidueInfo::one_letter_code)
-    .def_readonly("hydrogen_count", &ResidueInfo::hydrogen_count)
-    .def_readonly("weight", &ResidueInfo::weight)
-    .def("found", &ResidueInfo::found)
-    .def("is_standard", &ResidueInfo::is_standard)
-    .def("is_water", &ResidueInfo::is_water)
-    .def("is_nucleic_acid", &ResidueInfo::is_nucleic_acid)
-    .def("is_amino_acid", &ResidueInfo::is_amino_acid);
+  py::enum_<HowToNameCopiedChain>(m, "HowToNameCopiedChain")
+    .value("Short", HowToNameCopiedChain::Short)
+    .value("AddNumber", HowToNameCopiedChain::AddNumber)
+    .value("Dup", HowToNameCopiedChain::Dup)
+    ;
 
-  m.def("find_tabulated_residue", &find_tabulated_residue, py::arg("name"),
-        "Find chemical component information in the internal table.");
-  m.def("expand_protein_one_letter", &expand_protein_one_letter);
-  m.def("expand_protein_one_letter_string", &expand_protein_one_letter_string);
   m.def("one_letter_code",
         (std::string (*)(const std::vector<std::string>&)) &one_letter_code);
   m.def("one_letter_code",
         [](const ResidueSpan& span) { return one_letter_code(span); });
+  m.def("make_address", &make_address);
 
   py::enum_<CoorFormat>(m, "CoorFormat")
     .value("Unknown", CoorFormat::Unknown)
@@ -129,90 +79,14 @@ void add_mol(py::module& m) {
     .value("Mmjson", CoorFormat::Mmjson)
     .value("ChemComp", CoorFormat::ChemComp);
 
-  py::enum_<EntityType>(m, "EntityType")
-    .value("Unknown", EntityType::Unknown)
-    .value("Polymer", EntityType::Polymer)
-    .value("NonPolymer", EntityType::NonPolymer)
-    .value("Water", EntityType::Water);
-
-  py::enum_<PolymerType>(m, "PolymerType")
-    .value("PeptideL", PolymerType::PeptideL)
-    .value("PeptideD", PolymerType::PeptideD)
-    .value("Dna", PolymerType::Dna)
-    .value("Rna", PolymerType::Rna)
-    .value("DnaRnaHybrid", PolymerType::DnaRnaHybrid)
-    .value("SaccharideD", PolymerType::SaccharideD)
-    .value("SaccharideL", PolymerType::SaccharideL)
-    .value("Pna", PolymerType::Pna)
-    .value("CyclicPseudoPeptide", PolymerType::CyclicPseudoPeptide)
-    .value("Other", PolymerType::Other)
-    .value("Unknown", PolymerType::Unknown);
-
-  py::enum_<HowToNameCopiedChains>(m, "HowToNameCopiedChains")
-    .value("Short", HowToNameCopiedChains::Short)
-    .value("AddNumber", HowToNameCopiedChains::AddNumber)
-    .value("Dup", HowToNameCopiedChains::Dup)
-    ;
-
-  py::class_<Entity>(m, "Entity")
-    .def(py::init<std::string>())
-    .def_readwrite("name", &Entity::name)
-    .def_readwrite("subchains", &Entity::subchains)
-    .def_readwrite("entity_type", &Entity::entity_type)
-    .def_readwrite("polymer_type", &Entity::polymer_type)
-    .def_readwrite("full_sequence", &Entity::full_sequence)
-    .def_static("first_mon", &Entity::first_mon)
-    .def("__repr__", [](const Entity& self) { return tostr(self); });
-
-  py::class_<NcsOp>(m, "NcsOp")
-    .def(py::init<>())
-    .def_readwrite("id", &NcsOp::id)
-    .def_readwrite("given", &NcsOp::given)
-    .def_readonly("tr", &NcsOp::tr)
-    .def("apply", &NcsOp::apply)
-    .def("__repr__", [](const NcsOp& self) {
-        return tostr("<gemmi.NcsOp ", self.id,
-                     " |shift|=", self.tr.vec.length(),
-                     (self.given ? " (" : " (not "), "given)>");
-    });
-
-  py::class_<Assembly>(m, "Assembly")
-    .def_readonly("name", &Assembly::name)
-    .def_readonly("oligomeric_details", &Assembly::oligomeric_details)
-    .def("make_assembly", [](const Assembly& self, const Model& model,
-                             HowToNameCopiedChains how) {
-        return make_assembly(self, model, how, nullptr);
-    })
-    ;
-
-  py::enum_<Connection::Type>(m, "ConnectionType")
-    .value("Covale", Connection::Type::Covale)
-    .value("Disulf", Connection::Type::Disulf)
-    .value("Hydrog", Connection::Type::Hydrog)
-    .value("MetalC", Connection::Type::MetalC)
-    .value("Unknown", Connection::Type::Unknown);
-
-  py::class_<Connection>(m, "Connection")
-    .def(py::init<>())
-    .def_readwrite("name", &Connection::name)
-    .def_readwrite("link_id", &Connection::link_id)
-    .def_readwrite("type", &Connection::type)
-    .def_readwrite("asu", &Connection::asu)
-    .def_readwrite("partner1", &Connection::partner1)
-    .def_readwrite("partner2", &Connection::partner2)
-    .def_readwrite("reported_distance", &Connection::reported_distance)
-    .def("__repr__", [](const Connection& self) {
-        return tostr("<gemmi.Connection ", self.name, "  ",
-                     self.partner1.str(), " - ", self.partner2.str(), '>');
-    });
-
   py::bind_vector<std::vector<Connection>>(m, "ConnectionList");
   py::bind_vector<std::vector<NcsOp>>(m, "NcsOpList");
   py::bind_vector<std::vector<Entity>>(m, "EntityList");
-  py::bind_vector<std::vector<Assembly>>(m, "AssemblyList");
   py::bind_map<info_map_type>(m, "InfoMap");
 
-  py::class_<Structure>(m, "Structure")
+
+  py::class_<Structure> structure(m, "Structure");
+  structure
     .def(py::init<>())
     .def_readwrite("name", &Structure::name)
     .def_readwrite("cell", &Structure::cell)
@@ -223,6 +97,7 @@ void add_mol(py::module& m) {
     .def_readwrite("connections", &Structure::connections)
     .def_readwrite("assemblies", &Structure::assemblies)
     .def_readwrite("info", &Structure::info)
+    .def_readwrite("raw_remarks", &Structure::raw_remarks)
     .def("get_entity",
          (Entity* (Structure::*)(const std::string&)) &Structure::get_entity,
          py::arg("subchain"), py::return_value_policy::reference_internal)
@@ -253,40 +128,6 @@ void add_mol(py::module& m) {
     .def("merge_chain_parts", &Structure::merge_chain_parts,
          py::arg("min_sep")=0)
     .def("setup_cell_images", &Structure::setup_cell_images)
-    .def("make_pdb_headers", &make_pdb_headers)
-    .def("write_pdb", [](const Structure& st, const std::string& path,
-                         bool seqres_records, bool ssbond_records,
-                         bool link_records, bool cispep_records,
-                         bool ter_records, bool numbered_ter,
-                         bool ter_ignores_type, bool use_linkr) {
-       PdbWriteOptions options;
-       options.seqres_records = seqres_records;
-       options.ssbond_records = ssbond_records;
-       options.link_records = link_records;
-       options.cispep_records = cispep_records;
-       options.ter_records = ter_records;
-       options.numbered_ter = numbered_ter;
-       options.ter_ignores_type = ter_ignores_type;
-       options.use_linkr = use_linkr;
-       Ofstream f(path);
-       write_pdb(st, f.ref(), options);
-    }, py::arg("path"),
-       py::arg("seqres_records")=true, py::arg("ssbond_records")=true,
-       py::arg("link_records")=true, py::arg("cispep_records")=true,
-       py::arg("ter_records")=true, py::arg("numbered_ter")=true,
-       py::arg("ter_ignores_type")=false, py::arg("use_linkr")=false)
-    .def("write_minimal_pdb",
-         [](const Structure& st, const std::string& path) {
-       Ofstream f(path);
-       write_minimal_pdb(st, f.ref());
-    }, py::arg("path"))
-    .def("make_minimal_pdb", [](const Structure& st) -> std::string {
-       std::ostringstream os;
-       write_minimal_pdb(st, os);
-       return os.str();
-    })
-    .def("make_mmcif_document", &make_mmcif_document)
-    .def("make_mmcif_headers", &make_mmcif_headers)
     .def("add_entity_types", (void (*)(Structure&, bool)) &add_entity_types,
          py::arg("overwrite")=false)
     .def("assign_subchains", (void (*)(Structure&, bool)) &assign_subchains,
@@ -301,8 +142,8 @@ void add_mol(py::module& m) {
     .def("remove_ligands_and_waters",
          (void (*)(Structure&)) &remove_ligands_and_waters)
     .def("remove_empty_chains", (void (*)(Structure&)) &remove_empty_chains)
-    .def("assign_label_seq_id", &assign_label_seq_id, py::arg("force")=false)
     .def("shorten_chain_names", &shorten_chain_names)
+    .def("expand_ncs", &expand_ncs, py::arg("how"))
     .def("calculate_box", &calculate_box, py::arg("margin")=0.)
     .def("calculate_fractional_box", &calculate_fractional_box, py::arg("margin")=0.)
     .def("clone", [](const Structure& self) { return new Structure(self); })
@@ -310,22 +151,8 @@ void add_mol(py::module& m) {
         return tostr("<gemmi.Structure ", self.name, " with ",
                      self.models.size(), " model(s)>");
     });
-
-  pyAtomAddress
-    .def(py::init<>())
-    .def(py::init<const Chain&, const Residue&, const Atom&>())
-    .def(py::init<const std::string&, const SeqId&, const std::string&,
-                  const std::string&, char>(),
-         py::arg("chain"), py::arg("seqid"), py::arg("resname"),
-         py::arg("atom"), py::arg("altloc")='\0')
-    .def_readwrite("chain_name", &AtomAddress::chain_name)
-    .def_readwrite("res_id", &AtomAddress::res_id)
-    .def_readwrite("atom_name", &AtomAddress::atom_name)
-    .def_readwrite("altloc", &AtomAddress::altloc)
-    .def("__str__", &AtomAddress::str)
-    .def("__repr__", [](const AtomAddress& self) {
-        return tostr("<gemmi.AtomAddress ", self.str(), '>');
-    });
+    add_assign_label_seq_id(structure);
+    add_write(m, structure);
 
   py::class_<CRA>(m, "CRA")
     .def_readonly("chain", &CRA::chain)
@@ -367,9 +194,14 @@ void add_mol(py::module& m) {
          py::arg("name"), py::return_value_policy::reference_internal)
     .def("find_last_chain", &Model::find_last_chain,
          py::arg("name"), py::return_value_policy::reference_internal)
-    .def("add_chain", add_child<Model, Chain>,
-         py::arg("chain"), py::arg("pos")=-1,
-         py::return_value_policy::reference_internal)
+    .def("add_chain",
+         [](Model& self, const Chain& chain, int pos, bool unique_name) -> Chain& {
+           Chain& ref = add_child<Model, Chain>(self, chain, pos);
+           if (unique_name)
+             ensure_unique_chain_name(self, ref);
+           return ref;
+    }, py::arg("chain"), py::arg("pos")=-1, py::arg("unique_name")=false,
+       py::return_value_policy::reference_internal)
     // for compatibility with older gemmi version
     .def("add_chain", [](Model& self, const std::string& name) -> Chain& {
         self.chains.emplace_back(name);
@@ -428,13 +260,7 @@ void add_mol(py::module& m) {
     .def("__getitem__", &get_child<Chain, Residue>, py::arg("index"),
          py::return_value_policy::reference_internal)
     .def("__getitem__", [](Chain &ch, py::slice slice) -> py::list {
-        ssize_t start, stop, step, slength;
-        if (!slice.compute(ch.residues.size(), &start, &stop, &step, &slength))
-          throw py::error_already_set();
-        py::list l;
-        for (ssize_t i = 0; i < slength; ++i)
-          l.append(py::cast(&ch.residues[start + i * step]));
-        return l;
+        return getitem_slice(ch.residues, slice);
     }, py::return_value_policy::reference_internal)
     .def("__delitem__", remove_child<Chain>, py::arg("index"))
     .def("__delitem__", remove_children<Chain>)
@@ -542,26 +368,6 @@ void add_mol(py::module& m) {
                      self.size(), '>');
     });
 
-  pySeqId
-    .def(py::init<int, char>())
-    .def(py::init<const std::string&>())
-    .def_readwrite("num", &SeqId::num)
-    .def_readwrite("icode", &SeqId::icode)
-    .def("__str__", &SeqId::str)
-    .def("__repr__", [](const SeqId& self) {
-        return tostr("<gemmi.SeqId ", self.str(), '>');
-    });
-
-  pyResidueId
-    .def(py::init<>())
-    .def_readwrite("name", &ResidueId::name)
-    .def_readwrite("seqid", &ResidueId::seqid)
-    .def_readwrite("segment", &ResidueId::segment)
-    .def("__str__", &ResidueId::str)
-    .def("__repr__", [](const ResidueId& self) {
-        return tostr("<gemmi.ResidueId ", self.str(), '>');
-    });
-
   pyResidue
     .def(py::init<>())
     .def_readwrite("subchain", &Residue::subchain)
@@ -607,6 +413,12 @@ void add_mol(py::module& m) {
                      self.atoms.size(), " atoms>");
     });
 
+  py::enum_<CalcFlag>(m, "CalcFlag")
+    .value("NotSet", CalcFlag::NotSet)
+    .value("Determined", CalcFlag::Determined)
+    .value("Calculated", CalcFlag::Calculated)
+    .value("Dummy", CalcFlag::Dummy);
+
   pyAtom
     .def(py::init<>())
     .def_readwrite("name", &Atom::name)
@@ -618,7 +430,9 @@ void add_mol(py::module& m) {
     .def_readwrite("b_iso", &Atom::b_iso)
     .def_readwrite("aniso", &Atom::aniso)
     .def_readwrite("serial", &Atom::serial)
+    .def_readwrite("calc_flag", &Atom::calc_flag)
     .def_readwrite("flag", &Atom::flag)
+    .def_readwrite("tls_group_id", &Atom::tls_group_id)
     .def("is_hydrogen", &Atom::is_hydrogen)
     .def("has_altloc", &Atom::has_altloc)
     .def("b_eq", &Atom::b_eq)
@@ -636,19 +450,6 @@ void add_mol(py::module& m) {
         return r + buf;
     });
 
-  py::enum_<SupSelect>(m, "SupSelect")
-    .value("CaP", SupSelect::CaP)
-    .value("All", SupSelect::All);
-
-  py::class_<SupResult>(m, "SupResult")
-    .def_readonly("rmsd", &SupResult::rmsd)
-    .def_readonly("count", &SupResult::count)
-    .def_readonly("center1", &SupResult::center1)
-    .def_readonly("center2", &SupResult::center2)
-    .def_readonly("transform", &SupResult::transform)
-    .def("apply", &apply_superposition)
-    ;
-
   m.def("calculate_b_est", &calculate_b_est);
   m.def("calculate_angle", &calculate_angle,
         "Input: three points. Output: angle in radians.");
@@ -660,10 +461,9 @@ void add_mol(py::module& m) {
         py::arg("residue"), py::arg("next_residue"));
   m.def("calculate_sequence_weight", &calculate_sequence_weight,
         py::arg("sequence"), py::arg("unknown")=0.);
-  m.def("calculate_superposition",
-        [](const ResidueSpan& fixed, const ResidueSpan& movable,
-           PolymerType ptype, SupSelect sel, char altloc, bool current_rmsd) {
-          return calculate_superposition(fixed, movable, ptype, sel, altloc, current_rmsd);
-        }, py::arg("fixed"), py::arg("movable"), py::arg("ptype"), py::arg("sel"),
-           py::arg("altloc")='\0', py::arg("current_rmsd")=false);
+  m.def("make_assembly", [](const Assembly& assembly, const Model& model,
+                            HowToNameCopiedChain how) {
+        return make_assembly(assembly, model, how, nullptr);
+  })
+  ;
 }
