@@ -2,7 +2,7 @@
 
 #include "gemmi/symmetry.hpp"
 
-#include <pybind11/pybind11.h>
+#include "common.h"
 #include <pybind11/stl.h>
 #include <pybind11/stl_bind.h>
 #include <pybind11/numpy.h>
@@ -171,7 +171,7 @@ void add_symmetry(py::module& m) {
           throw std::domain_error("error: the size of the second dimension < 3");
         GroupOps gops = sg.operations();
         ReciprocalAsu asu(&sg);
-        for (ssize_t i = 0; i < h.shape(0); ++i) {
+        for (py::ssize_t i = 0; i < h.shape(0); ++i) {
           Op::Miller hkl = asu.to_asu({{h(i, 0), h(i, 1), h(i, 2)}}, gops).first;
           for (int j = 0; j != 3; ++j)
             h(i, j) = hkl[j];
